import os
import base64
import time
import tarfile

from kubernetes import config, client
from kubernetes.client import Configuration, V1DeleteOptions
from kubernetes.client.rest import ApiException
from kubernetes.client.api import core_v1_api
from kubernetes.stream import stream

from popper import utils as pu
from popper.cli import log as log
from popper.runner import StepRunner as StepRunner
from popper.runner_host import DockerRunner as HostDockerRunner


class KubernetesRunner(StepRunner):
    """Runs steps on a kubernetes cluster."""

    def __init__(self, **kw):
        super(KubernetesRunner, self).__init__(**kw)

        config.load_kube_config()

        c = Configuration()
        c.assert_hostname = False
        Configuration.set_default(c)
        self._kclient = core_v1_api.CoreV1Api()

        config.list_kube_config_contexts()

        self._namespace = self._config.resman_opts.get("namespace", "default")

        self._base_pod_name = pu.sanitized_name(f"pod", self._config.wid)
        self._base_pod_name = self._base_pod_name.replace("_", "-")

        self._init_pod_name = pu.sanitized_name("init-pod", self._config.wid)
        self._init_pod_name = self._init_pod_name.replace("_", "-")

        self._vol_claim_name = f"{self._base_pod_name}-pvc"
        self._vol_size = self._config.resman_opts.get("volume_size", "500Mi")

        self._init_pod_created = False
        self._vol_claim_created = False

    def __exit__(self, exc_type, exc_value, exc_traceback):
        super(KubernetesRunner, self).__exit__(exc_type, exc_value, exc_traceback)
        return True

    def run(self, step):
        """Execute a step in a kubernetes cluster."""
        self._pod_name = self._base_pod_name + f"-{step.id}"

        needs_build, _, img, tag, _ = self._get_build_info(step)

        if needs_build:
            log.fail(f"Cannot build ")

        image = f"{img}:{tag}"

        m = f"[{step.id}] kubernetes run {self._namespace}.{self._pod_name}"
        log.info(m)

        if self._config.dry_run:
            return 0

        ecode = 1
        try:
            if not self._vol_claim_created:
                if not self._vol_claim_exists():
                    self._vol_claim_create()
                self._vol_claim_created = True

            if not self._init_pod_created:
                e, self._pod_host_node = self._init_pod_schedule()
                if e:
                    raise Exception("None of the nodes are schedulable.")
                self._copy_ctx()
                self._init_pod_delete()
                self._init_pod_created = True

            self._pod_create(step, image, self._pod_host_node)
            self._pod_read_log()
            ecode = self._pod_exit_code()
        except Exception as e:
            log.fail(e)
        finally:
            self._pod_delete()

        log.debug(f"returning with {ecode}")
        return ecode

    def stop_running_tasks(self):
        """Delete the Pod and then the PersistentVolumeClaim upon receiving SIGINT.
        """
        log.debug("received SIGINT. deleting pod and volume claim")
        self._pod_delete()

    def _init_pod_schedule(self):
        """If a node selector is not provided, select a node randomly
        and stick to it."""
        e = 0
        pod_host_node = None

        if self._config.resman_opts.get("pod_host_node", None):
            e = self._init_pod_create(self._config.resman_opts.pod_host_node)
            pod_host_node = self._config.resman_opts.pod_host_node

        elif not self._config.resman_opts.get("persistent_volume_name", None):
            nodes = [
                node.metadata.labels["kubernetes.io/hostname"]
                for node in self._kclient.list_node().items
            ]
            for node in nodes:
                log.debug(f"trying to schedule init pod on {node}")
                e = self._init_pod_create(node)
                if not e:
                    pod_host_node = node
                    break
                else:
                    self._init_pod_delete()
        else:
            e = self._init_pod_create()
            pod_host_node = None

        return e, pod_host_node

    def _copy_ctx(self):
        """Tar up the workspace context and copy the tar file into
        the PersistentVolume in the Pod.
        """
        files = os.listdir(self._config.workspace_dir)
        with tarfile.open(
            os.path.join(self._config.workspace_dir, "ctx.tar.gz"), mode="w:gz"
        ) as archive:
            for f in files:
                archive.add(f)

        exec_command = ["/bin/sh"]
        response = stream(
            self._kclient.connect_get_namespaced_pod_exec,
            self._init_pod_name,
            self._namespace,
            command=exec_command,
            stderr=True,
            stdin=True,
            stdout=True,
            tty=False,
            _preload_content=False,
        )

        source_file = os.path.join(self._config.workspace_dir, "ctx.tar.gz")
        destination_file = "/workspace/ctx.tar.gz"

        with open(source_file, "rb") as image_file:
            encoded_string = base64.b64encode(image_file.read())

        commands = [
            f"echo {encoded_string.decode('utf-8')} | base64 --decode > {destination_file}"
        ]

        while response.is_open():
            response.update(timeout=1)
            if response.peek_stdout():
                log.debug(f"stdout: {response.read_stdout()}")
            if response.peek_stderr():
                log.debug(f"stderr: {response.read_stderr()}")
            if commands:
                c = commands.pop(0)
                log.debug(f"running command... {c}")
                response.write_stdin(c)
            else:
                break
        response.close()

        # extract the archive inside the pod
        exec_command = ["tar", "-zxvf", "/workspace/ctx.tar.gz"]
        response = stream(
            self._kclient.connect_get_namespaced_pod_exec,
            self._init_pod_name,
            self._namespace,
            command=exec_command,
            stderr=True,
            stdin=False,
            stdout=True,
            tty=False,
        )

        log.debug(response)

    def _init_pod_create(self, pod_host_node=None):
        """Create a init Pod mounted on a volume with alpine image so that 
        the `tar` utility is available by default and the workflow context 
        can be copied from the local machine into the volume.
        """
        ws_vol_mount = f"{self._init_pod_name}-ws"
        init_pod_conf = {
            "apiVersion": "v1",
            "kind": "Pod",
            "metadata": {"name": self._init_pod_name},
            "spec": {
                "restartPolicy": "Never",
                "containers": [
                    {
                        "image": "debian:stable",
                        "name": self._init_pod_name,
                        "workingDir": "/workspace",
                        "command": ["sleep", "infinity"],
                        "volumeMounts": [
                            {"name": ws_vol_mount, "mountPath": "/workspace",}
                        ],
                    }
                ],
                "volumes": [
                    {
                        "name": ws_vol_mount,
                        "persistentVolumeClaim": {"claimName": self._vol_claim_name,},
                    }
                ],
            },
        }

        if pod_host_node:
            init_pod_conf["spec"]["nodeSelector"] = {
                "kubernetes.io/hostname": pod_host_node
            }

        self._kclient.create_namespaced_pod(
            body=init_pod_conf, namespace=self._namespace
        )

        # loop and wait for the init pod to come up
        counter = 1
        while True:
            response = self._kclient.read_namespaced_pod(
                self._init_pod_name, namespace=self._namespace
            )
            if response.status.phase != "Pending":
                break

            log.debug(f"init pod {self._init_pod_name} not started yet")

            if counter == self._config.resman_opts.get("pod_retry_limit", 60):
                return 1

            time.sleep(1)
            counter += 1

        return 0

    def _init_pod_delete(self):
        """Teardown the init Pod after the context has been copied
        into the volume.
        """
        log.debug(f"deleting init pod {self._init_pod_name}")
        self._kclient.delete_namespaced_pod(
            self._init_pod_name, namespace=self._namespace, body=V1DeleteOptions()
        )

    def _vol_create(self, volume_name):
        """Create a default PersistentVolume of hostPath type.
        """
        vol_conf = {
            "kind": "PersistentVolume",
            "apiVersion": "v1",
            "metadata": {"name": volume_name, "labels": {"type": "host"},},
            "spec": {
                "persistentVolumeReclaimPolicy": "Recycle",
                "storageClassName": "manual",
                "capacity": {"storage": "1Gi",},
                "accessModes": ["ReadWriteMany"],
                "hostPath": {"path": "/tmp"},
            },
        }

        self._kclient.create_persistent_volume(body=vol_conf)

        counter = 1
        while True:
            response = self._kclient.read_persistent_volume(volume_name)
            if response.status.phase != "Pending":
                break

            log.debug(f"volume {volume_name} not created yet")

            if counter == 60:
                raise Exception("Timed out waiting for PersistentVolume creation")

            time.sleep(1)
            counter += 1

    def _vol_claim_create(self):
        """Create a PersistentVolumeClaim to claim usable storage space 
        from a previously created PersistentVolume.
        """
        if self._config.resman_opts.get("persistent_volume_name", None):
            volume_name = self._config.resman_opts.persistent_volume_name
        else:
            volume_name = f"pv-hostpath-popper-{self._config.wid}"
            if not self._vol_exists(volume_name):
                self._vol_create(volume_name)

        vol_claim_conf = {
            "apiVersion": "v1",
            "kind": "PersistentVolumeClaim",
            "metadata": {"name": self._vol_claim_name},
            "spec": {
                "storageClassName": "manual",
                "accessModes": ["ReadWriteMany"],
                "resources": {"requests": {"storage": self._vol_size}},
                "volumeName": volume_name,
            },
        }

        self._kclient.create_namespaced_persistent_volume_claim(
            namespace=self._namespace, body=vol_claim_conf
        )

        # wait for the volume claim to go into `Bound` state.
        counter = 1
        while True:
            response = self._kclient.read_namespaced_persistent_volume_claim(
                self._vol_claim_name, namespace=self._namespace
            )
            if response.status.phase != "Pending":
                break

            log.debug(f"volume claim {self._vol_claim_name} not created yet")

            if counter == 60:
                raise Exception("Timed out waiting for PersistentVolumeClaim creation")

            time.sleep(1)
            counter += 1

    def _vol_exists(self, volume_name):
        vol_exists = False
        try:
            self._kclient.read_persistent_volume(volume_name)
            vol_exists = True
        except ApiException as e:
            if e.reason != "Not Found":
                raise Exception(e)

        return vol_exists

    def _vol_claim_exists(self):
        vol_claim_exists = False
        try:
            self._kclient.read_namespaced_persistent_volume_claim(
                self._vol_claim_name, namespace=self._namespace
            )
            vol_claim_exists = True
        except ApiException as e:
            if e.reason != "Not Found":
                raise Exception(e)

        return vol_claim_exists

    def _vol_claim_delete(self):
        """Delete the PersistentVolumeClaim.
        """
        log.debug(f"deleting volume claim {self._vol_claim_name}")
        self._kclient.delete_namespaced_persistent_volume_claim(
            self._vol_claim_name, namespace=self._namespace, body=V1DeleteOptions()
        )

    def _pod_create(self, step, image, pod_host_node=None):
        """Start a Pod for each step.
        """
        log.debug(f"trying to start step pod on {pod_host_node}")
        env = self._prepare_environment(step)
        log.debug(env)

        ws_vol_mount = f"{self._pod_name}-ws"
        pod_conf = {
            "apiVersion": "v1",
            "kind": "Pod",
            "metadata": {"name": self._pod_name},
            "spec": {
                "restartPolicy": "Never",
                "containers": [
                    {
                        "image": image,
                        "name": f"{step.id}",
                        "workingDir": "/workspace",
                        "volumeMounts": [
                            {"name": ws_vol_mount, "mountPath": "/workspace",}
                        ],
                    }
                ],
                "volumes": [
                    {
                        "name": ws_vol_mount,
                        "persistentVolumeClaim": {"claimName": self._vol_claim_name,},
                    }
                ],
            },
        }

        if len(env.keys()) > 0:
            pod_conf["spec"]["containers"][0]["env"] = []
            for name, value in env.items():
                pod_conf["spec"]["containers"][0]["env"].append(
                    {"name": name, "value": value}
                )

        if pod_host_node:
            pod_conf["spec"]["nodeSelector"] = {"kubernetes.io/hostname": pod_host_node}

        runs = list(step.runs) if step.runs else None
        args = list(step.args) if step.args else None

        if runs:
            pod_conf["spec"]["containers"][0]["command"] = runs

        if args:
            pod_conf["spec"]["containers"][0]["args"] = args

        self._kclient.create_namespaced_pod(body=pod_conf, namespace=self._namespace)

        counter = 1
        while True:
            response = self._kclient.read_namespaced_pod(
                self._pod_name, namespace=self._namespace
            )
            if response.status.phase != "Pending":
                break

            log.debug(f"pod {self._pod_name} not started yet")

            if counter == self._config.resman_opts.get("pod_retry_limit", 60):
                raise Exception("Timed out waiting for Pod to start")

            time.sleep(1)
            counter += 1

    def _pod_read_log(self):
        """Read logs from the Pod after it moves into `Running` state.
        """
        log.debug(f"reading logs from {self._pod_name}")
        response = self._kclient.read_namespaced_pod_log(
            name=self._pod_name,
            namespace=self._namespace,
            follow=True,
            tail_lines=10,
            _preload_content=False,
        )
        for line in response:
            log.step_info(line.decode().rstrip())

    def _pod_exit_code(self):
        """Read the exit code from the Pod to decide the exit code of the step.
        """
        time.sleep(2)
        response = self._kclient.read_namespaced_pod(
            name=self._pod_name, namespace=self._namespace
        )
        log.debug(f"got status {response.status.phase}")
        if response.status.phase != "Succeeded":
            return 1
        return 0

    def _pod_delete(self):
        """Delete the Pod after it has Completed or Failed.
        """
        log.debug(f"deleting pod {self._pod_name}")
        self._kclient.delete_namespaced_pod(
            self._pod_name, namespace=self._namespace, body=V1DeleteOptions()
        )


class DockerRunner(KubernetesRunner, HostDockerRunner):
    """Runs steps on kubernetes; builds images locally using docker.
    """

    def __init__(self, **kw):
        super(DockerRunner, self).__init__(**kw)



import os

from setuptools import setup

version = {"__popper_version__": "0.0.0"}
version_file = "./popper/_version.py"
if os.path.isfile(version_file):
    with open(version_file) as f:
        exec(f.read(), version)

setup(
    name="popper",
    version=version["__popper_version__"],
    author="The Popper Development Team",
    author_email="ivotron@ucsc.edu",
    url="https://getpopper.io",
    description="Popper CLI tool to generate reproducible papers.",
    packages=["popper", "popper.commands"],
    include_package_data=True,
    install_requires=[
        "click==7.1.2",
        "docker==4.3.1",
        "dockerpty==0.4.1",
        "GitPython==3.1.7",
        "pykwalify==1.7.0",
        "python-box==5.1.1",
        "pyyaml==5.3.1",
        "spython==0.0.79",
        "kubernetes==11.0.0",
    ],
    extras_require={
        "dev": ["testfixtures==6.14.1", "black==19.10b0", "dunamai==1.3.0"]
    },
    entry_points="""
        [console_scripts]
        popper=popper.cli:cli
    """,
    project_urls={
        "Documentation": "https://popper.readthedocs.io",
        "Source": "https://github.com/getpopper/popper/",
        "Tracker": "https://github.com/getpopper/popper/issues",
    },
    keywords="popper reproducibility automation continuous-integration",
    license="MIT",
)



from click.testing import CliRunner
import click
from click import Context

from popper.commands import cmd_help
from .test_common import PopperTest
from popper.cli import PopperCLI, cmd_folder


class TestCommandHelp(PopperTest):
    def test_help(self):
        popper_cli = PopperCLI()
        subcommands = popper_cli.list_commands(click.Context(popper_cli))

        help_texts = []
        with self.assertLogs("popper") as test:
            runner = CliRunner()
            for subcommand in subcommands:
                result = runner.invoke(cmd_help.cli, [subcommand])
                self.assertEqual(result.exit_code, 0)
                target_command = popper_cli.get_command(
                    click.Context(popper_cli), subcommand
                )
                help_texts.append(target_command.get_help(click.Context(popper_cli)))

        for command_output, test_output in zip(help_texts, test.output):
            self.assertIn(
                command_output.replace(" ", "").replace("\n", ""),
                test_output.replace(" ", "").replace("\n", ""),
            )



import difflib
import importlib
import os
import signal

import click

import popper.log as logging

from popper import __version__

log = logging.setup_logging()
popper_version = __version__
cmd_folder = os.path.abspath(os.path.join(os.path.dirname(__file__), "commands"))


class Context(object):
    """Context Class.For more details refer
    https://click.palletsprojects.com/en/7.x/commands/#custom-multi-commands .
    """

    pass


pass_context = click.make_pass_decorator(Context, ensure=True)


class PopperCLI(click.MultiCommand):
    """Provides CLI interface for Popper."""

    def list_commands(self, ctx):
        """Returns the list of available commands in sorted order.

        Args:
          ctx(popper.cli.context): For process inter-command communication.
          For reference visit https://click.palletsprojects.com/en/7.x

        Returns:
          list: Returns the list of available commands.
        """
        rv = []
        for filename in os.listdir(cmd_folder):
            if filename.endswith(".py") and filename.startswith("cmd_"):
                rv.append(filename[4:-3])
        rv.sort()
        return rv

    def get_command(self, ctx, name):
        """Imports the command if available in commmands list and provides with
        most similar commands if the command is not present in the list.

        Args:
          name(str): The name of the command.

        Returns:
          click.core.Command: It is a new command and uses the decorated
            function as callback.For reference visit
            https://click.palletsprojects.com/en/7.x/api/#decorators .
        """
        cmd_path = f"popper.commands.cmd_{name}"
        cmd_spec = importlib.util.find_spec(cmd_path)

        if cmd_spec is not None:
            # found it
            cmd_mod = importlib.import_module(cmd_path)
            return cmd_mod.cli

        # given command doesn't exist, find a similarly named one
        commands = self.list_commands(ctx)
        similar_cmd = ", ".join(difflib.get_close_matches(name, commands, 1, 0.3))
        log.fail(
            f"Command '{name}' doesn't exist, did you mean '{similar_cmd}'? "
            "Run 'popper help' for more."
        )


@click.command(cls=PopperCLI)
@click.version_option(__version__, message=f"Popper version {popper_version}")
@pass_context
def cli(ctx):
    """Popper command line interface."""
    from popper.runner import WorkflowRunner

    signal.signal(signal.SIGINT, WorkflowRunner.signal_handler)
    signal.signal(signal.SIGUSR1, WorkflowRunner.signal_handler)



import importlib
import os

__version__ = "0.0.0"

_repo_dir = os.path.join(os.path.dirname(os.path.realpath(__file__)), "..", "..")
_version_file = os.path.join(os.path.dirname(os.path.realpath(__file__)), "_version.py")

# check if dunamai is available
_dunamai_spec = importlib.util.find_spec("dunamai")
_dunamai_found = _dunamai_spec is not None

if _dunamai_found:

    # if dunamai is found, then we use it to display the version
    import dunamai

    # if codebase is inside a git repo, define GIT_DIR so dunamai gets the info for the
    # popper repo instead of wherever popper is being invoked from
    _git_dir = os.path.join(_repo_dir, ".git")

    if os.path.isdir(_git_dir):
        os.environ["GIT_DIR"] = _git_dir

    try:
        __version__ = dunamai.Version.from_git().serialize()
        _ver = f"__popper_version__ = '{__version__}'"
    except RuntimeError as e:
        # only raise if not a dunamai-raised error
        if "This does not appear to be a Git project" not in str(e):
            raise e

    with open(_version_file, "w") as v:
        v.write(_ver + "\n")

    # unset GIT_DIR
    os.environ.pop("GIT_DIR", None)

elif os.path.isfile(_version_file):
    # codebase not in a popper repo, and version file exists, so let's read from it
    from popper._version import __popper_version__ as _ver

    __version__ = _ver
